/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#ifdef OPTION_CRYPTO_SUPPORT

#include "common.h"
#include "crypto.h"

// CRYPT_OPENSSL should never be defined. It's here just to make it 
// easy to reuse the openssl version later

#ifdef CRYPT_OPENSSL // use the openssl crypto if it was specified
int crypt_block_blowfish(u64 insize, u64 *outsize, u8 *inbuf, u8 *outbuf, u8 *password, int passlen, int enc)
{
	EVP_CIPHER_CTX ctx;
	u8 iv[] = "fsarchiv";
	int cursize=0;
	int outlen=0;

	// init cipher structures
	EVP_CIPHER_CTX_init(&ctx);
	EVP_CipherInit_ex(&ctx, EVP_bf_cfb(), NULL, NULL, NULL, enc);
	EVP_CIPHER_CTX_set_key_length(&ctx, passlen);
	EVP_CipherInit_ex(&ctx, NULL, NULL, password, iv, enc);
	
	if(EVP_CipherUpdate(&ctx, outbuf+cursize, &outlen, inbuf, insize)!=1)
	{	errprintf("EVP_CipherUpdate() failed\n");
		EVP_CIPHER_CTX_cleanup(&ctx);
		return 1;
	}
	cursize+=outlen;
	if(EVP_CipherFinal_ex(&ctx, outbuf+cursize, &outlen)!=1)
	{	errprintf("EVP_CipherFinal_ex() failed\n");
		EVP_CIPHER_CTX_cleanup(&ctx);
		return 1;
	}
	cursize+=outlen;
	
	EVP_CIPHER_CTX_cleanup(&ctx);
	
	*outsize=cursize;
	
	return 0;
}
#else // CRYPT_OPENSSL // use the libgcrpyt crypto by default
int crypt_block_blowfish(u64 insize, u64 *outsize, u8 *inbuf, u8 *outbuf, u8 *password, int passlen, int enc)
{
	gcry_cipher_hd_t hd;
	u8 iv[] = "fsarchiv";
	int res;
	
	// init
	if ((password==NULL) || (passlen==0))
		return -1;
	
	// gcry_error_t gcry_cipher_open (gcry_cipher_hd_t *handle, int algo, int mode, unsigned int flags);
	if (gcry_cipher_open(&hd, GCRY_CIPHER_BLOWFISH, GCRY_CIPHER_MODE_CFB, GCRY_CIPHER_SECURE))
	{
		errprintf("gcry_cipher_open() failed\n");
		gcry_cipher_close(hd);
		return -1;
	}
	
	// gcry_error_t gcry_cipher_setkey (gcry_cipher_hd_t hd, const void *key, size_t keylen);
	if (gcry_cipher_setkey(hd, password, passlen))
	{
		errprintf("gcry_cipher_setkey() failed\n");
		gcry_cipher_close(hd);
		return -1;
	}
	
	// gcry_error_t gcry_cipher_setiv (gcry_cipher_hd_t h, const void *k, size_t l)
	if (gcry_cipher_setiv(hd, iv, strlen((char*)iv)))
	{
		errprintf("gcry_cipher_setiv() failed\n");
		gcry_cipher_close(hd);
		return -1;
	}

	switch(enc)
	{
		case 1: // encrypt
			// gcry_error_t gcry_cipher_encrypt (gcry_cipher_hd_t h, void *out, size_t outsize, const void *in, size_t inlen);
			res=gcry_cipher_encrypt(hd, outbuf, insize, inbuf, insize);
			break;
		case 0: // decrypt
			// gcry_error_t gcry_cipher_decrypt (gcry_cipher_hd_t h, void *out, size_t outsize, const void *in, size_t inlen);
			res=gcry_cipher_decrypt(hd, outbuf, insize, inbuf, insize);
			break;
		default: // invalid
			errprintf("invalid parameter: enc=%d\n", (int)enc);
			gcry_cipher_close(hd);
			return -1;
	}
	
	// void gcry_md_close (gcry_md_hd_t hd);	
	gcry_cipher_close(hd);
	
	*outsize=insize;
	return (res==0)?(0):(-1);
}
#endif // CRYPT_OPENSSL

#endif // OPTION_CRYPTO_SUPPORT

int generate_random_buffer(u8 *buf, int bufsize)
{
	memset(buf, 0, bufsize);
#ifdef OPTION_CRYPTO_SUPPORT
#ifdef CRYPT_OPENSSL
	int ret=RAND_pseudo_bytes(buf, bufsize);
#else // CRYPT_OPENSSL
	gcry_randomize(buf, bufsize, GCRY_STRONG_RANDOM);
#endif // CRYPT_OPENSSL
#else // OPTION_CRYPTO_SUPPORT
	int i;
	for (i=0; i < bufsize; i++)
		buf[i]=rand()%256;
#endif // OPTION_CRYPTO_SUPPORT
	return 0;
}
